from datetime import datetime

from sklearn.model_selection import train_test_split
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.metrics import f1_score
import tensorflow as tf

from wellcomeml.ml.attention import HierarchicalAttention

TENSORBOARD_LOG_DIR = "logs/scalars/" + datetime.now().strftime("%Y%m%d-%H%M%S")
METRIC_DICT = {
    'precision': tf.keras.metrics.Precision(name='precision'),
    'recall': tf.keras.metrics.Recall(name='recall'),
    'auc': tf.keras.metrics.AUC(name='auc')
}
CALLBACK_DICT = {
    'tensorboard': tf.keras.callbacks.TensorBoard(log_dir=TENSORBOARD_LOG_DIR)
}


class BiLSTMClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        learning_rate=0.01,
        batch_size=32,
        nb_epochs=5,
        dropout=0.1,
        nb_layers=2,
        hidden_size=100,
        dense_size=20,
        l2=1e-6,
        attention_heads='same',
        multilabel=False,
        attention=False,
        metrics=["precision", "recall"],
        callbacks=["tensorboard"],
        feature_approach="max"
    ):
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.nb_epochs = nb_epochs
        self.dropout = dropout
        self.nb_layers = nb_layers
        self.hidden_size = hidden_size
        self.l2 = l2
        self.attention_heads = attention_heads
        self.dense_size = dense_size
        self.multilabel = multilabel
        self.attention = attention
        self.metrics = metrics
        self.callbacks = callbacks
        self.feature_approach = feature_approach

    def _build_model(self, sequence_length, vocab_size, nb_outputs,
                     embedding_matrix=None, metrics=["precision", "recall"]):
        output_activation = (
            "sigmoid" if nb_outputs == 1 or self.multilabel else "softmax"
        )
        embedding_size = embedding_matrix.shape[1] if embedding_matrix else self.hidden_size

        def residual_bilstm(x1, l2):
            x2 = tf.keras.layers.Bidirectional(
                tf.keras.layers.LSTM(
                    int(x1.shape[-1] / 2), return_sequences=True, kernel_regularizer=l2
                )
            )(x1)
            return tf.keras.layers.add([x1, x2])

        def residual_attention(x1):
            x2 = HierarchicalAttention(self.attention_heads)(x1)
            x2 = tf.keras.layers.Dropout(self.dropout)(x2)
            x2 = tf.keras.layers.LayerNormalization()(x2)
            if self.attention_heads == 'same':
                return tf.keras.layers.add([x1, x2])
            else:
                return x2

        l2 = tf.keras.regularizers.l2(self.l2)
        embeddings_initializer = (
            tf.keras.initializers.Constant(embedding_matrix)
            if embedding_matrix
            else "uniform"
        )
        inp = tf.keras.layers.Input(shape=(sequence_length,))
        x = tf.keras.layers.Embedding(
            vocab_size,
            embedding_size,
            input_length=sequence_length,
            embeddings_initializer=embeddings_initializer,
        )(inp)
        x = tf.keras.layers.Dropout(
            self.dropout,
            noise_shape=(None, sequence_length, 1))(x)
        for _ in range(self.nb_layers):
            x = residual_bilstm(x, l2)
        if self.attention:
            x = residual_attention(x)
        if self.feature_approach == 'max':
            x = tf.keras.layers.GlobalMaxPooling1D()(x)
        elif self.feature_approach == 'sum':
            x = tf.keras.backend.sum(x, axis=1)
        elif self.feature_approach == 'concat':
            x = tf.concat(x, axis=1)
        else:
            raise NotImplementedError
        x = tf.keras.layers.Dense(self.dense_size, kernel_regularizer=l2)(x)
        out = tf.keras.layers.Dense(
            nb_outputs, activation=output_activation, kernel_regularizer=l2
        )(x)
        model = tf.keras.Model(inp, out)

        optimizer = tf.keras.optimizers.Adam(lr=self.learning_rate, clipnorm=1.0)
        metrics = [
            METRIC_DICT[m] if m in METRIC_DICT else m
            for m in metrics
        ]
        model.compile(optimizer=optimizer, loss="binary_crossentropy", metrics=metrics)
        return model

    def fit(self, X, Y, embedding_matrix=None, *_):
        sequence_length = X.shape[1]
        vocab_size = X.max() + 1

        nb_outputs = Y.max() if not self.multilabel else Y.shape[1]

        self.model = self._build_model(sequence_length, vocab_size, nb_outputs,
                                       embedding_matrix, self.metrics)

        X_train, X_val, Y_train, Y_val = train_test_split(
            X, Y, test_size=0.1, shuffle=True
        )
        callbacks = [
            CALLBACK_DICT[c] if c in CALLBACK_DICT else c
            for c in self.callbacks
        ]
        self.model.fit(
            X_train,
            Y_train,
            epochs=self.nb_epochs,
            batch_size=self.batch_size,
            validation_data=(X_val, Y_val),
            callbacks=callbacks,
        )
        return self

    def predict(self, X, *_):
        return self.model(X).numpy() > 0.5

    def score(self, X, Y):
        Y_pred = self.predict(X)
        return f1_score(Y, Y_pred, average="micro")



#!/usr/bin/env python3
# coding: utf-8

"""
A generic vectorizer that can fallback to tdidf or bag of words from sklearn
or embed using bert, doc2vec etc
"""

from sklearn.base import BaseEstimator, TransformerMixin
from wellcomeml.ml.frequency_vectorizer import WellcomeTfidf
from wellcomeml.ml.bert_vectorizer import BertVectorizer
from wellcomeml.ml.keras_vectorizer import KerasVectorizer
from wellcomeml.ml.doc2vec_vectorizer import Doc2VecVectorizer


class Vectorizer(BaseEstimator, TransformerMixin):
    """
    Abstract class, sklearn-compatible, that can vectorize texts using
    various models.

    """

    def __init__(self, embedding="tf-idf", cache_transformed=False, **kwargs):
        """
        Args:
            embedding(str): One of `['bert', 'tf-idf']`
            cache_transformed(bool): Caches the last transformed vector X (
            useful if performing Grid-search as part of a pipeline)
        """
        self.embedding = embedding
        self.cache_transformed = cache_transformed

        vectorizer_dispatcher = {
            "tf-idf": WellcomeTfidf,
            "bert": BertVectorizer,
            "keras": KerasVectorizer,
            "doc2vec": Doc2VecVectorizer,
        }

        if not vectorizer_dispatcher.get(embedding):
            raise ValueError(f"Model {embedding} not available")

        self.vectorizer = vectorizer_dispatcher.get(embedding)(**kwargs)

    def fit(self, X=None, *_):
        return self.vectorizer.fit(X)

    def transform(self, X, *_):
        X_transformed = self.vectorizer.transform(X)

        if self.cache_transformed:
            self.X_transformed = X_transformed

        return X_transformed

    def fit_transform(self, X, y=None, *_):
        # Slightly modified fit_transform so it can work with the
        # cache_transformed
        self.fit(X)
        return self.transform(X)

    def save_transformed(self, path, X_transformed):
        """
        Saves transformed vector X_transformed vector, using the corresponding
        save_transformed method for the specific vectorizer.

        Args:
            path: A path to the embedding file
            X_transformed: A transformed vector (as output by using the
            .transform method)

        """
        save_method = getattr(self.vectorizer.__class__, "save_transformed", None)
        if not save_method:
            raise NotImplementedError(
                f"Method save_transformed not implemented"
                f" for class "
                f"{self.vectorizer.__class__.__name__}"
            )

        return save_method(path=path, X_transformed=X_transformed)

    def load_transformed(self, path):
        """
        Loads transformed vector X_transformed vector, using the corresponding
        load method for the specific vectorizer.

        Args:
            path: A path to the file containing embedded vectors

        Returns:
            X_transformed (array), like the one returned by the the
            fit_transform function.
        """
        load_method = getattr(self.vectorizer.__class__, "load_transformed", None)
        if not load_method:
            raise NotImplementedError(
                f"Method load_transformed not implemented"
                f" for class "
                f"{self.vectorizer.__class__.__name__}"
            )

        return load_method(path=path)



import subprocess
import sys


def download(download_target):
    if download_target == "models":
        subprocess.run([
            'python', '-m', 'spacy', 'download', 'en_core_web_sm'])
    elif download_target == "deeplearning-models":
        subprocess.run([
            'python', '-m', 'spacy', 'download', 'en_trf_bertbaseuncased_lg'])
    elif download_target == "non_pypi_packages":
        subprocess.run([
            'pip', 'install', 'git+https://github.com/epfml/sent2vec.git'])
    else:
        print(f"{download_target} is not one of models,deeplearning-models")


if __name__ == '__main__':
    command = sys.argv.pop(1)
    if command != "download":
        print("Only available command is download")
        exit()

    download_target = sys.argv.pop(1)
    download(download_target)



from collections import defaultdict
import logging
import os

import numpy as np
from sklearn.base import ClusterMixin
from sklearn.cluster import DBSCAN, OPTICS, KMeans
from sklearn.manifold import TSNE
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import silhouette_score
from sklearn.pipeline import Pipeline
from hdbscan import HDBSCAN
import umap

from wellcomeml.ml import vectorizer
from wellcomeml.logger import logger

CACHE_DIR = os.path.expanduser("~/.cache/wellcomeml")


class TextClustering(object):
    """
    Class

    Attributes:
        vectorizer: The embedding Vectorizer object
        reducer: A dimensionality reduction object
        clustering: A clustering model object
        cluster_ids: Ids of the cluster
        cluster_names: Names of the clusters
        cluster_kws: Keywords for the clusters (only if embedding=tf-idf)
    """
    def __init__(self, embedding='tf-idf', reducer='umap', clustering='dbscan',
                 cluster_reduced=True, n_kw=10, params={}):
        """
        Initialises a cluster pipeline

        Args:
            embedding(str): 'tf-idf', 'bert', 'keras' or 'doc2vec'.
            reducer(str): 'umap' or 'tsne'
            clustering(str): 'dbscan', 'kmeans', 'optics', 'hdbscan' or any
            class that is a sklearn.base.ClusterMixin
            cluster_reduced(bool): Whether clustering will be applied after
            reducing the datapoints with reducer, or straight after embedding
            n_kw(int): If n_kw > 0 and embedding='tf-idf', the model
            outputs, besides a cluster id, a list of keywords for each
            cluster.
            params(dict): Dictionary containing extra parameters for
            embedding, reducer or clustering models. Example:
            {'clustering': {'eps': 0.1}, 'embedding': {'ngram_range': (1,2)}}
        """
        self.embedding = embedding
        self.vectorizer = vectorizer.Vectorizer(embedding=embedding,
                                                **params.get('embedding', {}))
        self.n_kw = n_kw
        self.cluster_reduced = cluster_reduced

        reducer_dispatcher = {
            'umap': umap.UMAP,
            'tsne': TSNE
        }
        self.reducer = reducer_dispatcher[reducer](**params.get('reducer', {}))

        clustering_dispatcher = {
            'dbscan': DBSCAN,
            'kmeans': KMeans,
            'optics': OPTICS,
            'hdbscan': HDBSCAN
        }

        if clustering in clustering_dispatcher:
            self.clustering = clustering_dispatcher[clustering](
                **params.get('clustering', {})
            )
        elif isinstance(clustering, ClusterMixin):
            self.clustering = clustering(**params.get('clustering', {}))
        else:
            raise ValueError('clustering has to be one of the available '
                             'clusters or a sklearn ClusterMixin.')
        self.cluster_ids = None
        self.cluster_names = None
        self.cluster_kws = None
        self.silhouette = None

    def fit(self, X, *_):
        """
        Fits all clusters in the pipeline

        Args:
            X: A list of texts

        Returns:
            A TextClustering object

        """
        self._fit_step(X, step='vectorizer')
        self._fit_step(step='reducer')
        self._fit_step(step='clustering')

        if self.embedding == 'tf-idf' and self.n_kw:
            self._find_keywords(self.embedded_points.toarray(), n_kw=self.n_kw)

        return self

    def _fit_step(self, X=None, step='vectorizer'):
        """Internal function for partial fitting only a certain step"""
        if step == 'vectorizer':
            self.embedded_points = self.vectorizer.fit_transform(X)
        elif step == 'reducer':
            self.reduced_points = \
                self.reducer.fit_transform(self.embedded_points)
        elif step == 'clustering':
            points = (
                self.reduced_points if self.cluster_reduced else
                self.embedded_points
            )
            self.clustering.fit(points)
            self.cluster_ids = self.clustering.labels_

    def optimise(self, X, param_grid, n_cluster_range=None, max_noise=0.2,
                 verbose=False):
        """
        Optimises the clustering silhouette based on a parameter grid,
        a range on number of clusters and a range on noise.

        It is customised to avoid re-fitting of intermediate
        steps (vectorizer and reducer) more than necessary.

        Args:
            X: A list of texts to be clustered
            param_grid: A parameter grid, example:

            param_grid = {'reducer': {'min_dist': [0.0, 0.2], 'n_neighbors': [
            2,3,5], 'metric': ['cosine', 'euclidean']},
            'clustering': {'min_samples': [2, 5], 'eps': [0.5, 1, 1.5]}}

            n_cluster_range (2-uple of ints): A 2-uple describing the
            max and min number of clusters (e.g.: (10, 20)). If unset,
            will just choose the best silhopuette

            max_noise (float in [0,1]): The maximum fraction of points
            unclustered. Default: 0.2

        Returns:
            A dictionary containing "params_list", "silhouette" (the
            silhouete for each parameter) and "best_clustering" (the best
            clustering parameters)

        """

        min_n_clusters = (n_cluster_range[0] if n_cluster_range else 0)
        max_n_clusters = (n_cluster_range[1] if n_cluster_range else 10**5)

        # Linearises Dictionary to be compatible with grid search so it
        # becomes one dictionary with 'step__parameter'
        if self.cluster_reduced:
            pipeline = Pipeline([('vectorizer', self.vectorizer),
                                 ('reducer', self.reducer),
                                 ('clustering', self.clustering)],
                                memory=CACHE_DIR)
            params = {
                **{f'reducer__{key}': value for key, value in
                   param_grid.get('reducer', {}).items()}
            }
        else:
            self.vectorizer.cache_transformed = True
            pipeline = Pipeline([('vectorizer', self.vectorizer),
                                 ('clustering', self.clustering)],
                                memory=CACHE_DIR)
            params = {}

        params = {
            **params,
            **{f'vectorizer__{key}': value for key, value in
               param_grid.get('vectorizer', {}).items()}
        }

        params = {
            **params,
            **{f'clustering__{key}': value for key, value in
               param_grid.get('clustering', {}).items()}
        }

        grid = GridSearchCV(
            estimator=pipeline,
            param_grid=params,
            scoring={
                'silhouette': _clustering_score,
                'noise': _clustering_noise,
                'n_clusters': _number_of_clusters
            },
            refit='silhouette'
        )

        logging_level = logger.level
        if verbose <= 1:
            # Previously disable logging to allow the loading bar to run
            # uninterruptly. Will reset after.
            logging.getLogger().setLevel(logging.WARNING)
            logger.setLevel(logging.WARNING)

        # Prunes result to actually optimise under constraints
        best_silhouette = -1000
        best_params = {}
        grid.fit(X, y=None)
        for params, silhouette, noise, n_clusters in zip(
                grid.cv_results_['params'],
                grid.cv_results_['mean_test_silhouette'],
                grid.cv_results_['mean_test_noise'],
                grid.cv_results_['mean_test_n_clusters']
        ):
            if min_n_clusters <= n_clusters <= max_n_clusters\
                    and noise <= max_noise\
                    and silhouette > best_silhouette:
                best_silhouette = silhouette
                best_params = params

        self.silhouette = best_silhouette

        self.set_params(best_params, from_parameter_grid=True)
        # Fits the pipeline again with the best parameters
        self.fit(X)

        logger.setLevel(logging_level)

        return best_params

    def stability(self):
        """Function to calculate how stable the clusters are"""
        raise NotImplementedError

    def break_cluster(self, cluster_id):
        """Breaks a specific by re-clustering a subset"""
        raise NotImplementedError

    def merge_clusters(self, dictionary):
        """Merge cluster based on ids"""
        self.cluster_ids = [dictionary.get(cluster_id, cluster_id) for
                            cluster_id in self.cluster_ids]

    def set_params(self, params, from_parameter_grid=False):
        """
        Set parameters of clustering, reducer and vectorizer

        Args:
            params: A dictionary of parameters
            from_parameter_grid: Whether the dictionary is in the form of
            sklearn parameter grid (e.g. {'clustering__eps': 1}), or a two
            level dictionary (e.g. {'clustering': {'eps': 1}}) (default).

        Returns:
            None

        """
        if from_parameter_grid:
            # This is to convert from a sklearn parameter to a two-level
            # dictionary, which is what we work with
            new_params = defaultdict(dict)

            for key, value in params.items():
                level_1_key, level_2_key = key.split('__')
                new_params[level_1_key][level_2_key] = value

            params = new_params

        self.vectorizer.set_params(**params.get('vectorizer', {}))
        self.reducer.set_params(**params.get('reducer', {}))
        self.clustering.set_params(**params.get('clustering', {}))

    def _find_keywords(self, X_transformed, n_kw=10):
        if self.embedding != 'tf-idf':
            raise ValueError(f'Cannot find keyword for '
                             f'{self.embedding} vectorizer')
        idx_to_word = {value: key for key, value in
                       self.vectorizer.vectorizer.vocabulary_.items()}

        # Initialise a dictionary to calculate the aggregated vector for each
        # cluster

        aggregated = {key: np.zeros(X_transformed.shape[1])
                      for key in self.cluster_ids}

        for i in range(len(X_transformed)):
            aggregated[self.cluster_ids[i]] += X_transformed[i]

        kw_dictionary = {
            key: ','.join(_find_words_from_frequency(vector,
                                                     idx_to_word,
                                                     n_kw=n_kw))
            for key, vector in aggregated.items()
        }

        self.cluster_kws = [
            kw_dictionary.get(cluster_id) for cluster_id in self.cluster_ids
        ]


def _find_words_from_frequency(vector, idx_to_word, n_kw=5):
    """
    Given a vector of frequencies and a dictionary connecting indexes
     to word, find most prominent words
    up to n_kw, eliminating possible duplicates or sub-words.

    Args:
        vector: A vector representing frequencies (or any other scores) for
        each word.
        idx_to_word: A dictionary mapping indices in `vector` to words.
        n_kw: number of "prominent" words to output.

    Returns:
        A list of n_kw words (strings).

    """
    sorted_idx = np.argsort(vector)[::-1]
    # Tests if new words do not overlap with existing ones (this is to
    # exclude things like: 'cluster,clustering,clustered')

    current_kw_list = []
    i = 0

    while len(current_kw_list) < n_kw and i <= len(vector)-1:
        new_word = idx_to_word[sorted_idx[i]]
        used_word = False
        for kw in current_kw_list:
            if new_word in kw or kw in new_word:
                used_word = True
                break

        if not used_word:
            current_kw_list += [new_word]

        i += 1

    return current_kw_list


def _number_of_clusters(estimator, X, y=None):
    """Utility to return number of clusters, excluding noise (if exists)"""
    label_set = set(estimator['clustering'].labels_)
    return len(label_set)-1*(-1 in label_set)


def _clustering_noise(estimator, X, y=None):
    """Returns the clustering noise (predictions = -1)"""
    return len([x for x in estimator['clustering'].labels_ if x == -1])


def _clustering_score(estimator, X, y=None):
    """
    Scores a clustering based on the silhouette score

    Args:
        estimator: A sklearn Pipeline. Assumes the estimator has a set
        called 'embedding' and a step called 'clustering', which is a
        sklearn.base.ClusterMixin

        X: A vector of size n x m, where n is the number of points and m
         the dimension of each point

    Returns:
        The average silhouette score of X, given the estimator labels,
         or 0 if only one cluster has been found

    """
    if hasattr(estimator.named_steps, 'reducer'):
        X_points = estimator.named_steps['reducer'].embedding_
    elif hasattr(estimator.named_steps['vectorizer'], 'X_transformed'):
        X_points = estimator.named_steps['vectorizer'].X_transformed
    else:
        X_points = estimator['vectorizer'].transform(X)

    predicted_labels = estimator['clustering'].labels_

    return silhouette_score(X_points, predicted_labels) \
        if 2 <= len(np.unique(predicted_labels)) <= len(predicted_labels)-1 \
        else 0



import tensorflow as tf


class SelfAttention(tf.keras.layers.Layer):
    """https://papers.nips.cc/paper/7181-attention-is-all-you-need.pdf"""

    def __init__(self, attention_dim=20):
        super(SelfAttention, self).__init__()
        self.attention_dim = attention_dim

    def build(self, input_shape):
        self.WQ = self.add_weight(
            shape=(input_shape[-1], self.attention_dim),
            trainable=True,
            initializer="uniform",
        )
        self.WK = self.add_weight(
            shape=(input_shape[-1], self.attention_dim),
            trainable=True,
            initializer="uniform",
        )
        self.WV = self.add_weight(
            shape=(input_shape[-1], input_shape[-1]),
            trainable=True,
            initializer="uniform",
        )

    def call(self, X):
        """
        In: (batch_size, sequence_length, embedding_dimension)
        Out: (batch_size, sequence_length, embedding_dimension)
        """
        Q = tf.matmul(X, self.WQ)
        K = tf.matmul(X, self.WK)
        V = tf.matmul(X, self.WV)

        attention_scores = tf.nn.softmax(tf.matmul(Q, tf.transpose(K, perm=[0, 2, 1])))
        return tf.matmul(attention_scores, V)


class FeedForwardAttention(tf.keras.layers.Layer):
    """https://colinraffel.com/publications/iclr2016feed.pdf"""

    def __init__(self):
        super(FeedForwardAttention, self).__init__()

    def build(self, input_shape):
        self.W = self.add_weight(
            shape=(input_shape[-1], 1), trainable=True, initializer="uniform"
        )

    def call(self, X):
        """
        In: (batch_size, sequence_length, embedding_dimension)
        Out: (batch_size, embedding_dimension)
        """
        e = tf.math.tanh(tf.matmul(X, self.W))
        attention_scores = tf.nn.softmax(e)
        return tf.matmul(tf.transpose(X, perm=[0, 2, 1]), attention_scores)


class HierarchicalAttention(tf.keras.layers.Layer):
    """https://www.aclweb.org/anthology/N16-1174/"""

    def __init__(self, attention_heads='same'):
        super(HierarchicalAttention, self).__init__()
        self.attention_heads = attention_heads

    def build(self, input_shape):
        if self.attention_heads == 'same':
            nb_attention_heads = input_shape[-2]
        else:
            nb_attention_heads = self.attention_heads
        self.attention_matrix = self.add_weight(
            shape=(input_shape[-1], nb_attention_heads),
            trainable=True,
            initializer="uniform",
        )

    def call(self, X):
        """
        In: (batch_size, sequence_length, embedding_dimension)
        Out: (batch_size, sequence_length, embedding_dimension)
        """
        attention_scores = tf.nn.softmax(
            tf.math.tanh(tf.matmul(X, self.attention_matrix))
        )
        return tf.matmul(tf.transpose(attention_scores, perm=[0, 2, 1]), X)



import os
import random

from wellcomeml.datasets.download import check_cache_and_download


def _load_data_spacy(data_path, inc_outside=True):
    """
    Load data in Spacy format:
    X = list of sentences (plural) / documents ['the cat ...', 'some dog...', ...]
    Y = list of list of entity tags for each sentence
        [[{'start': 36, 'end': 46, 'label': 'PERSON'}, {..}, ..], ... ]
    inc_outside = False: don't include none-entities in the output

    Raw format is:
    '-DOCSTART- -X- O O\n\nEU NNP I-NP I-ORG\nrejects VBZ I-VP O\nGerman JJ I-NP I-MISC...'
    where each article is separated by '-DOCSTART- -X- O O\n',
    each sentence is separate by a blank line,
    and the entity information is in the form
    'EU NNP I-NP I-ORG' (A word, a part-of-speech (POS) tag,
        a syntactic chunk tag and the named entity tag)
    """

    X = []
    Y = []
    with open(data_path) as f:
        articles = f.read().split("-DOCSTART- -X- O O\n\n")
        articles = articles[1:]  # The first will be blank
        for article in articles:
            # Each sentence in the article is separated by a blank line
            sentences = article.split("\n\n")
            for sentence in sentences:
                char_i = 0  # A counter for the entity start and end character indices
                sentence_text = ""
                sentence_tags = []
                entities = sentence.split("\n")
                for entity in entities:
                    # Due to the splitting on '\n' sometimes we are left with empty elements
                    if len(entity) != 0:
                        token, _, _, tag = entity.split(" ")
                        sentence_text += token + " "
                        if tag != "O" or inc_outside:
                            sentence_tags.append(
                                {
                                    "start": char_i,
                                    "end": char_i + len(token),
                                    "label": tag,
                                }
                            )
                        char_i += len(token) + 1  # plus 1 for the space separating
                if sentence_tags != []:
                    X.append(sentence_text)
                    Y.append(sentence_tags)

    return X, Y


def load_conll(split="train", shuffle=True, inc_outside=True):
    path = check_cache_and_download("conll")

    if split == "train":
        train_data_path = os.path.join(path, "eng.train")
        X, Y = _load_data_spacy(train_data_path, inc_outside=inc_outside)
    elif split == "test":
        test_data_path = os.path.join(path, "eng.testa")
        X, Y = _load_data_spacy(test_data_path, inc_outside=inc_outside)
    elif split == "evaluate":
        eval_data_path = os.path.join(path, "eng.testb")
        X, Y = _load_data_spacy(eval_data_path, inc_outside=inc_outside)
    else:
        raise ValueError(
            f"Split argument {split} is not one of train, test or evaluate"
        )

    if shuffle:
        data = list(zip(X, Y))
        random.shuffle(data)
        X, Y = zip(*data)

    return X, Y



from collections import defaultdict
import os

import tensorflow as tf
from transformers import BertConfig, BertTokenizer, TFBertForSequenceClassification

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import train_test_split
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError

from wellcomeml.ml.keras_utils import CategoricalMetrics, MetricMiniBatchHistory


class SemanticEquivalenceClassifier(BaseEstimator, TransformerMixin):
    """
    Class to fine-tune BERT-type models for semantic equivalence, for example
    paraphrase, textual similarity and other NLU tasks
    """

    def __init__(
        self,
        pretrained="bert",
        batch_size=32,
        eval_batch_size=32 * 2,
        learning_rate=3e-5,
        test_size=0.2,
        max_length=128,
    ):
        """

        Args:
            pretrained(str): 'bert' or 'scibert'
            batch_size(int): The batch size for training
            eval_batch_size(int): The evaluation batch size
            learning_rate(float): Learning rate (usually between 0.1 and 1e-5)
            test_size(float): Proportion of data used for testing (default 0.2)
            max_length: Maximum length of text in characters.
             Zero pads every text smaller than this number and cuts out
             any text bigger than that number
        """
        self.pretrained = pretrained
        self.batch_size = batch_size
        self.eval_batch_size = eval_batch_size
        self.learning_rate = learning_rate
        self.test_size = test_size
        self.max_length = max_length

        # Defines attributes that will be initialised later
        self.config = None
        self.tokenizer = None
        self.model = None
        self.train_dataset = None
        self.valid_dataset = None
        self.train_steps = None
        self.valid_steps = None

        self.history = defaultdict(list)

    # Bert models have a tensorflow checkopoint, otherwise,
    # we need to load the pytorch versions with the parameter `from_pt=True`

    def _initialise_models(self):
        if self.pretrained == "bert":
            model_name = "bert-base-cased"
            from_pt = False
        elif self.pretrained == "scibert":
            model_name = "allenai/scibert_scivocab_cased"
            from_pt = True

        self.config = BertConfig.from_pretrained(model_name, num_labels=2)
        self.tokenizer = BertTokenizer.from_pretrained(model_name)
        self.model = TFBertForSequenceClassification.from_pretrained(
            model_name, config=self.config, from_pt=from_pt
        )
        return self.model

    def _prep_dataset_generator(self, X, y):
        features = ["input_ids", "attention_mask", "token_type_ids"]

        batch_encoding = self.tokenizer.batch_encode_plus(
            X, max_length=self.max_length, add_special_tokens=True,
        )

        def gen_train():
            for i in range(len(X)):
                features = {
                    k: pad(batch_encoding[k][i], self.max_length)
                    for k in batch_encoding
                }

                yield (features, int(y[i]))

        input_element_types = ({feature: tf.int32 for feature in features}, tf.int64)
        input_element_tensors = (
            {feature: tf.TensorShape([None]) for feature in features},
            tf.TensorShape([]),
        )

        dataset = tf.data.Dataset.from_generator(
            gen_train, input_element_types, input_element_tensors,
        )

        return dataset

    def _compile_model(self, metrics=[]):
        opt = tf.keras.optimizers.Adam(learning_rate=self.learning_rate, epsilon=1e-08)

        loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)

        precision = CategoricalMetrics(metric="precision")
        recall = CategoricalMetrics(metric="recall")
        f1_score = CategoricalMetrics(metric="f1_score")
        accuracy = tf.keras.metrics.SparseCategoricalAccuracy("accuracy")

        metrics = [accuracy, precision, recall, f1_score] + metrics

        self.model.compile(optimizer=opt, loss=loss, metrics=metrics)

    def _tokenize(self, X):
        return self.tokenizer.batch_encode_plus(
            X,
            max_length=self.max_length,
            add_special_tokens=True,
            pad_to_max_length=True,
            return_tensors="tf",
        )

    def _prep_data_for_prediction(self, X):
        X_tokenized = self._tokenize(X)
        return tf.convert_to_tensor(self.model.predict(X_tokenized))

    def fit(self, X, y, random_state=None, epochs=3, metrics=[], **kwargs):
        """
        Fits a sentence similarity model

        Args:
            X: List of 2-uples: [('Sentence 1', 'Sentence 2'), ....]
            y: list of 0 (not similar) and 1s (similar)
            random_state: a random state for the train_test_split
            epochs: number_of epochs

        Returns:
            A fitted model

        """
        # Initialises/downloads model if not trained before.
        # If trained, fits extra epochs without the transformations
        try:
            check_is_fitted(self)
        except NotFittedError:
            self._initialise_models()

            # Train/val split
            X_train, X_valid, y_train, y_valid = train_test_split(
                X, y, test_size=self.test_size, random_state=random_state
            )

            # Generates tensorflow dataset
            self.train_dataset = self._prep_dataset_generator(X_train, y_train)
            self.valid_dataset = self._prep_dataset_generator(X_valid, y_valid)

            # Generates mini-batches and stores in a class variable
            self.train_dataset = (
                self.train_dataset.shuffle(self.max_length)
                .batch(self.batch_size)
                .repeat(-1)
            )
            self.valid_dataset = self.valid_dataset.batch(self.eval_batch_size)

            self._compile_model()

            self.train_steps = len(X_train) // self.batch_size
            self.valid_steps = len(X_valid) // self.eval_batch_size
        finally:
            history = self.model.fit(
                self.train_dataset,
                epochs=epochs,
                steps_per_epoch=self.train_steps,
                validation_data=self.valid_dataset,
                validation_steps=self.valid_steps,
                callbacks=[MetricMiniBatchHistory()],
                **kwargs
            )

            # Accumulates history of metrics from different partial fits
            for metric, values in history.history.items():
                self.history[metric] += values

            self.trained_ = True

        return self

    def score(self, X):
        """
        Calculates scores for model prediction

        Args:
            X: List of 2-uples: [('Sentence 1', 'Sentence 2'), ....]

        Returns:
            An array of shape len(X) x 2 with scores for classes 0 and 1

        """
        predictions = self._prep_data_for_prediction(X)

        return tf.keras.activations.softmax(predictions).numpy()

    def predict(self, X):
        """
        Calculates the predicted class (0 - for negative and 1 for positive)
        for X.

        Args:
            X: List of 2-uples: [('Sentence 1', 'Sentence 2'), ....]

        Returns:
            An array of 0s and 1s

        """
        return self.score(X).argmax(axis=1)

    def save(self, path):
        """Saves model to path"""
        os.makedirs(path, exist_ok=True)
        self.model.save_pretrained(path)

    def load(self, path):
        """Loads model from path"""
        self._initialise_models()
        self.model = TFBertForSequenceClassification.from_pretrained(path)
        self.trained_ = True


class SemanticEquivalenceMetaClassifier(SemanticEquivalenceClassifier):
    """
    Class to fine tune Semantic Classifier, allowing the possibility of
    adding metadata to the classification. Extends
    SemanticEquivalenceClassifier, and extends modules for data prep
    and model initialisation. Fit, predict, score remain intact
    """

    def __init__(self, n_numerical_features, dropout_rate=0.1, **kwargs):
        """
        Initialises SemanticEquivalenceMetaClassifier, with n_numerical_features

        Args:
            n_numerical_features(int): Number of features of the model which
            are not text
            dropout_rate(float): Dropout rate after concatenating features
            (default = 0.1)
            **kwargs: Any kwarg to `SemanticEquivalenceClassifier`
        """
        super().__init__(**kwargs)
        self.n_numerical_features = n_numerical_features
        self.dropout_rate = dropout_rate
        self.model = None

    def _separate_features(self, X):
        X_text = [x[:2] for x in X]
        X_numerical = [x[2: 2 + self.n_numerical_features] for x in X]

        return X_text, X_numerical

    def _initialise_models(self):
        """Extends/overrides super class initialisation of model"""
        super_model = super()._initialise_models()

        # Define text input features
        text_features = ["input_ids", "attention_mask", "token_type_ids"]
        input_text_tensors = [
            tf.keras.layers.Input(
                name=feature_name, shape=tf.TensorShape([None]), dtype=tf.int32
            )
            for feature_name in text_features
        ]

        input_numerical_data_tensor = [
            tf.keras.layers.Input(
                name="numerical_metadata",
                shape=tf.TensorShape([self.n_numerical_features]),
                dtype=tf.float32,
            )
        ]
        # Calls the CLS layer of Bert
        x = super_model.bert(input_text_tensors)[1]

        # Drop out layer to the Bert features
        x = super_model.dropout(x, training=False)

        # Concatenates with numerical features
        x = tf.keras.layers.concatenate(
            [x, input_numerical_data_tensor[0]], name="concatenate"
        )

        # Dense layer that will be used for softmax prediction later
        x = tf.keras.layers.Dense(2, name="dense_layer")(x)

        self.model = tf.keras.Model(input_text_tensors + input_numerical_data_tensor, x)

    def _prep_dataset_generator(self, X, y):
        """Overrides/extends the super class data preparation"""
        text_features = ["input_ids", "attention_mask", "token_type_ids"]
        X_text, X_numerical = self._separate_features(X)

        batch_encoding_text = self.tokenizer.batch_encode_plus(
            X_text, max_length=self.max_length, add_special_tokens=True,
        )

        def gen_train():
            for i in range(len(X)):
                features = {
                    k: pad(batch_encoding_text[k][i], self.max_length)
                    for k in batch_encoding_text
                }
                features["numerical_metadata"] = X_numerical[i]

                yield (features, int(y[i]))

        input_element_types = (
            {
                **{feature: tf.int32 for feature in text_features},
                **{"numerical_metadata": tf.float32},
            },
            tf.int64,
        )
        input_element_tensors = (
            {
                **{feature: tf.TensorShape([None]) for feature in text_features},
                **{"numerical_metadata": tf.TensorShape([self.n_numerical_features])},
            },
            tf.TensorShape([]),
        )

        dataset = tf.data.Dataset.from_generator(
            gen_train, input_element_types, input_element_tensors,
        )

        return dataset

    def _prep_data_for_prediction(self, X):
        X_text, X_numerical = self._separate_features(X)

        X_processed = self._tokenize(X_text)
        X_processed["numerical_metadata"] = tf.convert_to_tensor(X_numerical)

        predictions = tf.convert_to_tensor(self.model.predict(X_processed))
        return predictions

    def fit(self, X, y, **kwargs):
        """
        Fits semantic classifier

        Args:
            X(list of t-uples): Assumes that each element x of X has length
            2+n_numerical_features, the first two components are texts and
            remaining components are numerical features (e.g.)
            y: list of 0 (not similar) and 1s (similar)
            random_state: a random state for the train_test_split
            epochs: number_of epochs

        Returns:
             Array of probabilities, of shape len(X) x 2

        """
        return super().fit(X, y, **kwargs)

    def score(self, X):
        """
        Calculates scores for metadata semantic classifier.

        Args:
            X(list of t-uples): Assumes that each element x of X has length
            2+n_numerical_features, the first two components are texts and
            remaining components are numerical features (e.g.)

        Returns:
            Numpy array of probabilities, of shape len(X) x 2

        """

        return super().score(X)

    def predict(self, X):
        """
        Calculates scores for metadata semantic classifier.

        Args:
            X(list of t-uples): Assumes that each element x of X has length
            2+n_numerical_features, the first two components are texts and
            remaining components are numerical features (e.g.)

        Returns:
            Array of predictions of length len(X)

        """
        return super().predict(X)

    def save(self, path):
        """Saves meta model to path"""
        self.model.save(path)

    def load(self, path):
        """Loads metamodel from path"""
        self._initialise_models()
        self.model = tf.keras.models.load_model(path)
        self.trained_ = True


def pad(x, pad_len):
    """Old versions of transformers do not pad by default"""
    return x + [0] * (pad_len - len(x))



"""
Voting classifier with ability to accept pretrained estimators
and return results for multilabel Y

Some of the code is taken from https://gist.github.com/tomquisel/a421235422fdf6b51ec2ccc5e3dee1b4
"""
from sklearn.utils.validation import check_is_fitted
from sklearn.ensemble import VotingClassifier
from sklearn.exceptions import NotFittedError
import numpy as np


class WellcomeVotingClassifier(VotingClassifier):
    def __init__(self, multilabel=False, *args, **kwargs):
        super(WellcomeVotingClassifier, self).__init__(*args, **kwargs)
        self.pretrained = self._is_pretrained()
        self.multilabel = multilabel

    def _is_pretrained(self):
        try:
            check_is_fitted(self, 'estimators')
        except NotFittedError:
            return False
        return True

    def _get_estimators(self):
        if type(self.estimators) == list:
            return [est for est in self.estimators]
        else:  # tuple with named estimators
            return [est for _, est in self.estimators]

    def predict(self, X):
        if self.pretrained:
            check_is_fitted(self, 'estimators')

            estimators = self._get_estimators()

            if self.voting == "soft":
                Y_probs = np.array([est.predict_proba(X) for est in estimators])
                Y_prob = np.mean(Y_probs, axis=0)
                if self.multilabel:
                    return np.array(Y_prob > 0.5, dtype=int)
                else:
                    return np.argmax(Y_prob, axis=1)
            else:  # hard voting
                Y_preds = [est.predict(X) for est in estimators]
                if self.multilabel:
                    Y_preds = np.array(Y_preds)
                    axis = 0
                else:
                    Y_preds = np.column_stack(Y_preds)
                    axis = 1
                return np.apply_along_axis(
                    lambda x: np.argmax(np.bincount(x)),
                    axis=axis,
                    arr=Y_preds)
        else:
            return super(WellcomeVotingClassifier, self).predict(X)



import random
import csv
import os

from wellcomeml.datasets.download import check_cache_and_download


def load_split(data_path):
    X = []
    Y = []
    with open(data_path) as f:
        csvreader = csv.DictReader(f, delimiter="\t")
        for line in csvreader:
            X.append(line["sentence"])
            Y.append(line["labels"].split(","))
    return X, Y


def load_hoc(split="train", shuffle=True):
    path = check_cache_and_download("hoc")

    if split == "train":
        train_data_path = os.path.join(path, "train.tsv")
        X, Y = load_split(train_data_path)
    elif split == "test":
        test_data_path = os.path.join(path, "test.tsv")
        X, Y = load_split(test_data_path)
    else:
        raise ValueError(f"Split argument {split} is not one of train or test")

    if shuffle:
        data = list(zip(X, Y))
        random.shuffle(data)
        X, Y = zip(*data)

    return X, Y



#!/usr/bin/env python3
# coding: utf-8

import spacy


class SpacyDocToProdigy:
    """Convert spacy documents into prodigy format
    """

    def run(self, docs):
        """
        Cycle through docs and return prodigy docs.
        """

        return list(self.return_one_prodigy_doc(doc) for doc in docs)

    def return_one_prodigy_doc(self, doc):
        """Given one spacy document, yield a prodigy style dict

        Args:
            doc (spacy.tokens.doc.Doc): A spacy document

        Returns:
            dict: Prodigy style document

        """

        if not isinstance(doc, spacy.tokens.doc.Doc):
            raise TypeError("doc must be of type spacy.tokens.doc.Doc")

        text = doc.text
        spans = []
        tokens = []

        for token in doc:
            tokens.append({
                "text": token.text,
                "start": token.idx,
                "end": token.idx + len(token.text),
                "id": token.i,
            })

        for ent in doc.ents:
            spans.append({
                "token_start": ent.start,
                "token_end": ent.end,
                "start": ent.start_char,
                "end": ent.end_char,
                "label": ent.label_,
            })

        out = {
            "text": text,
            "spans": spans,
            "tokens": tokens,
        }

        return out



__name__ = "wellcomeml"
__version__ = "2020.9.0"
__description__ = """Utilities for managing nlp models and for
                     processing text-related data at Wellcome Data Labs"""
__url__ = "https://github.com/wellcometrust/wellcomeml/tree/master"
__author__ = "Wellcome Trust DataLabs Team"
__author_email__ = "Grp_datalabs-datascience@Wellcomecloud.onmicrosoft.com"
__license__ = "MIT"



#!/usr/bin/env python
# coding: utf8
"""
Train a convolutional neural network for multilabel classification
of grants
Adapted from https://github.com/explosion/spaCy/blob/master/examples/training/train_textcat.py
"""
from spacy_transformers import (
    TransformersLanguage,
    TransformersWordPiecer,
    TransformersTok2Vec,
)
from spacy_transformers.model_registry import (
    register_model,
    get_last_hidden,
    flatten_add_lengths,
)
from spacy.util import minibatch, compounding
from spacy._ml import zero_init, logistic
from thinc.t2v import Pooling, mean_pool
from thinc.v2v import Affine
from thinc.api import chain
from sklearn.metrics import precision_recall_fscore_support, f1_score
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import train_test_split
import numpy as np
import spacy
import torch

import random
import time

from wellcomeml.logger import logger


@register_model("sigmoid_last_hidden")
def sigmoid_last_hidden(nr_class, *, exclusive_classes=False, **cfg):
    width = cfg["token_vector_width"]
    return chain(
        get_last_hidden,
        flatten_add_lengths,
        Pooling(mean_pool),
        zero_init(Affine(nr_class, width, drop_factor=0.0)),
        logistic,
    )


is_using_gpu = spacy.prefer_gpu()
if is_using_gpu:
    torch.set_default_tensor_type("torch.cuda.FloatTensor")


class BertClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        threshold=0.5,
        n_iterations=5,
        pretrained="bert",
        batch_size=8,
        learning_rate=1e-5,
        dropout=0.1,
        l2=1e-4,
        validation_split=0.1,
    ):
        self.threshold = threshold
        self.batch_size = batch_size
        self.dropout = dropout
        self.learning_rate = learning_rate
        self.n_iterations = n_iterations
        self.pretrained = pretrained
        self.l2 = l2
        self.validation_split = validation_split

    def _init_nlp(self):
        if self.pretrained == "bert":
            self.nlp = spacy.load("en_trf_bertbaseuncased_lg")
        elif self.pretrained == "scibert":
            name = "scibert-scivocab-uncased"
            path = "models/scibert_scivocab_uncased"

            nlp = TransformersLanguage(trf_name=name, meta={"lang": "en"})
            nlp.add_pipe(nlp.create_pipe("sentencizer"))
            nlp.add_pipe(TransformersWordPiecer.from_pretrained(nlp.vocab, path))
            nlp.add_pipe(TransformersTok2Vec.from_pretrained(nlp.vocab, path))
            self.nlp = nlp
        else:
            logger.info(f"{self.pretrained} is not among bert, scibert")
            raise
        # TODO: Add a parameter for exclusive classes, non multilabel scenario
        self.textcat = self.nlp.create_pipe(
            "trf_textcat",
            config={"exclusive_classes": False, "architecture": "sigmoid_last_hidden"},
        )

        self.nlp.add_pipe(self.textcat, last=True)

        for label in self.unique_labels:
            self.textcat.add_label(label)

    def load(self, model_dir):
        self.nlp = spacy.load(model_dir)
        # another hack to get the labels from textcat since
        # spacy does not serialise every attribute
        for pipe_name, pipe in self.nlp.pipeline:
            # to cover case of textcat and trf_textcat
            if "textcat" in pipe_name:
                self.unique_labels = pipe.labels

    def save(self, output_dir):
        self.nlp.to_disk(output_dir)

    def _label_binarizer_inverse_transform(self, Y_train):
        "Transforms Y matrix to labels which are the non zero indices"
        data = []
        for row in Y_train:
            row_data = [str(i) for i, item in enumerate(row) if item]
            data.append(row_data)
        # Do we need to convert to numpy?
        return np.array(data)

    def fit(self, X, Y):
        X_train, X_test, Y_train, Y_test = train_test_split(
            X, Y, random_state=42, test_size=self.validation_split
        )
        self.unique_labels = [str(i) for i in range(Y_train.shape[1])]
        self._init_nlp()
        n_iter = self.n_iterations
        train_texts = X_train
        train_tags = self._label_binarizer_inverse_transform(Y_train)

        train_cats = [
            {label: label in tags for label in self.unique_labels}
            for tags in train_tags
        ]

        train_data = list(zip(train_texts, [{"cats": cats} for cats in train_cats]))

        other_pipes = [pipe for pipe in self.nlp.pipe_names if "trf" not in pipe]
        with self.nlp.disable_pipes(*other_pipes):  # only train textcat
            optimizer = self.nlp.resume_training()
            optimizer.alpha = self.learning_rate
            optimizer.L2 = self.l2
            logger.info("Training the model...")
            logger.info(
                "{:^5}\t{:^5}\t{:^5}\t{:^5}\t{:^5}\t{:^5}".format(
                    "ITER", "LOSS", "P", "R", "F", "TF"
                )
            )
            batch_sizes = compounding(4.0, self.batch_size, 1.001)
            # dropout = decaying(0.6, 0.2, 1e-4)
            self.losses = []
            for i in range(n_iter):
                start_epoch = time.time()
                nb_examples = 0
                losses = {}
                # batch up the examples using spaCy's minibatch
                random.shuffle(train_data)
                batches = minibatch(train_data, size=batch_sizes)
                for batch in batches:
                    texts, annotations = zip(*batch)
                    next_dropout = self.dropout  # next(dropout)
                    self.nlp.update(
                        texts,
                        annotations,
                        sgd=optimizer,
                        drop=next_dropout,
                        losses=losses,
                    )
                    nb_examples += len(texts)
                with self.textcat.model.use_params(optimizer.averages):
                    Y_test_pred = self.predict(X_test)
                    p, r, f1, _ = precision_recall_fscore_support(
                        Y_test, Y_test_pred, average="micro"
                    )
                loss = losses["trf_textcat"]
                self.losses.append(loss)

                epoch_seconds = time.time() - start_epoch
                speed = nb_examples / epoch_seconds
                logger.info(
                    "{0:5d}\t{1:.3f}\t{2:.3f}\t{3:.3f}\t{4:.3f}\t{5:.2f}".format(
                        i, loss, p, r, f1, speed
                    )
                )
        return self

    def partial_fit(self, X, Y, classes=None):
        if not hasattr(self, "unique_labels"):
            self.unique_labels = [str(i) for i in range(Y.shape[1])]
            self._init_nlp()
            self.losses = []

        texts = X

        train_tags = self._label_binarizer_inverse_transform(Y)
        train_cats = [
            {label: label in tags for label in self.unique_labels}
            for tags in train_tags
        ]
        annotations = [{"cats": cats} for cats in train_cats]

        other_pipes = [pipe for pipe in self.nlp.pipe_names if "trf" not in pipe]
        with self.nlp.disable_pipes(*other_pipes):  # only train textcat
            if not hasattr(self, "optimizer"):
                self.optimizer = self.nlp.resume_training()
                self.optimizer.alpha = self.learning_rate

            losses = {}
            self.nlp.update(
                texts, annotations, sgd=self.optimizer, drop=self.dropout, losses=losses
            )
            self.losses.append(losses["trf_textcat"])
        return self

    def predict(self, X):
        def binarize_output(doc):
            cats = doc.cats
            out = [
                1 if cats[label] > self.threshold else 0 for label in self.unique_labels
            ]
            return out
        doc_gen = self.nlp.pipe(X)
        return np.array([binarize_output(doc) for doc in doc_gen])

    def predict_proba(self, X):
        def get_proba(doc):
            cats = doc.cats
            out = [cats[label] for label in self.unique_labels]
            return out
        doc_gen = self.nlp.pipe(X)
        return np.array([get_proba(doc) for doc in doc_gen])

    def score(self, X, Y):
        Y_pred = self.predict(X)
        return f1_score(Y_pred, Y, average="micro")



"""
Creates a knowledge base using the vocab from an NLP model
and pretrains the entity encodings using the entity descriptions

See https://spacy.io/usage/training#entity-linker for where I got this code from
"""
from pathlib import Path
import subprocess
import os

from spacy.vocab import Vocab
from spacy.kb import KnowledgeBase

# bin is also a spaCy package
from bin.wiki_entity_linking.train_descriptions import EntityEncoder
import spacy


class SpacyKnowledgeBase(object):
    def __init__(
        self, kb_model="en_core_web_lg", desc_width=64, input_dim=300, num_epochs=5
    ):
        """
        Input:
            kb_model: spacy pretrained model with word embeddings
            desc_width: length of entity vectors
            input_dim: dimension of pretrained input vectors
            num_epochs: number of epochs in training entity encodings
        """
        self.kb_model = kb_model
        self.desc_width = desc_width
        self.input_dim = input_dim
        self.num_epochs = num_epochs

    def train(self, entities, list_aliases):
        """
        entities: a dict of each entity, it's description and it's corpus frequency
        list_aliases: a list of dicts for each entity e.g.
            [{
                'alias':'Farrar',
                'entities': ['Q1', 'Q2'],
                'probabilities': [0.4, 0.6]
            }]
            probabilities are 'prior probabilities' and must sum to < 1
        """
        try:
            nlp = spacy.load(self.kb_model)
        except IOError:
            subprocess.run(["python", "-m", "spacy", "download", self.kb_model])
            # pkg_resources need to be reloaded to pick up the newly installed models
            import pkg_resources
            import imp

            imp.reload(pkg_resources)
            nlp = spacy.load(self.kb_model)

        print("Loaded model '%s'" % self.kb_model)
        kb = KnowledgeBase(vocab=nlp.vocab, entity_vector_length=self.desc_width)

        # set up the data
        entity_ids = []
        descriptions = []
        freqs = []
        for key, value in entities.items():
            desc, freq = value
            entity_ids.append(key)
            descriptions.append(desc)
            freqs.append(freq)

        # training entity description encodings
        # this part can easily be replaced with a custom entity encoder
        encoder = EntityEncoder(
            nlp=nlp,
            input_dim=self.input_dim,
            desc_width=self.desc_width,
            epochs=self.num_epochs,
        )

        encoder.train(description_list=descriptions, to_print=True)

        # get the pretrained entity vectors
        embeddings = encoder.apply_encoder(descriptions)

        # set the entities, can also be done by calling `kb.add_entity` for each entity
        kb.set_entities(entity_list=entity_ids, freq_list=freqs, vector_list=embeddings)

        # adding aliases, the entities need to be defined in the KB beforehand
        for alias in list_aliases:
            kb.add_alias(
                alias=alias["alias"],
                entities=alias["entities"],
                probabilities=alias["probabilities"],
            )
        self.kb = kb
        return self.kb

    def save(self, output_dir):
        output_dir = Path(output_dir)
        if not output_dir.exists():
            output_dir.mkdir()
        kb_path = os.path.join(output_dir, "kb")
        self.kb.dump(kb_path)
        print("Saved KB to", kb_path)

        vocab_path = os.path.join(output_dir, "vocab")
        self.kb.vocab.to_disk(vocab_path)
        print("Saved vocab to", vocab_path)

    def load(self, output_dir):
        kb_path = os.path.join(output_dir, "kb")
        vocab_path = os.path.join(output_dir, "vocab")
        print("Loading vocab from", vocab_path)
        print("Loading KB from", kb_path)
        vocab = Vocab().from_disk(vocab_path)
        kb = KnowledgeBase(vocab=vocab)
        kb.load_bulk(kb_path)
        self.kb = kb
        return self.kb

    def __str__(self):
        print(self.kb.get_size_entities(), "kb entities:", self.kb.get_entity_strings())
        print(self.kb.get_size_aliases(), "kb aliases:", self.kb.get_alias_strings())



from .__version__ import (
    __name__, __version__, __description__, __url__,
    __author__, __author_email__, __license__
)
from .logger import logger

__all__ = [
    __name__, __version__, __description__, __url__,
    __author__, __author_email__, __license__, logger
]



"""
Vectorizer that exposes sklearn interface to sent2vec
paper and codebase. https://github.com/epfml/sent2vec
"""
from sklearn.base import TransformerMixin, BaseEstimator

from wellcomeml.utils import check_cache_and_download


class Sent2VecVectorizer(BaseEstimator, TransformerMixin):
    def __init__(self, pretrained=None):
        self.pretrained = pretrained

    def fit(self, *_):

        try:
            import sent2vec
        except ImportError:
            from wellcomeml.__main__ import download

            download("non_pypi_packages")
            import sent2vec

        if self.pretrained:
            model_path = check_cache_and_download(self.pretrained)
            self.model = sent2vec.Sent2vecModel()
            self.model.load_model(model_path)
        else:
            # Custom training not yet implemented
            raise NotImplementedError(
                "Fit only implemented for loading pretrained models"
            )
        return self

    def transform(self, X):
        return self.model.embed_sentences(X)



# WiNER: A Wikipedia Annotated Corpus for Named Entity Recognition
# https://www.aclweb.org/anthology/I17-1042/

import tarfile
import os
from tqdm import tqdm
import random

from wellcomeml.datasets.download import check_cache_and_download
from wellcomeml.logger import logger


def yield_article_entities(f):
    for i, line in enumerate(f):
        line = line.decode("utf-8").replace("\n", "")  # because line is of type bytes
        if line.startswith("ID "):
            article_id = line.split("ID ")[1]
        else:
            entities_data = line.replace("\t", " ")
            sentence_id, begin, end, ent_type = entities_data.split(" ")
            # The data is stored as strings, but they are all numerical
            # and sentence id = 3 refers to the 2nd sentence in this article
            # so for querying the list of sentences in an article
            # later it is useful to have this as an integer
            yield (article_id, int(sentence_id), int(begin), int(end), int(ent_type))


def yield_merged_entities(entities):
    # Yield successive spans of entities into one
    # e.g.
    # [{'start': 0, 'end': 2, 'label': '3-B'},
    # {'start': 3, 'end': 5, 'label': '3-E'}]
    # -> {'start': 0, 'end': 5, 'label': '3'}
    merged_entity = None
    for i, tag in enumerate(entities):
        if tag["label"][-2:] == "-B":
            # Yield the previous entity if it existed
            if merged_entity:
                yield merged_entity
            # Start creating a new merged entity
            merged_entity = tag.copy()
            merged_entity["label"] = tag["label"][0]
            if i + 1 == len(entities):
                yield merged_entity
        elif tag["label"][-2:] == "-E":
            if merged_entity:
                # merged_entity should always exist if you reach a -E, but
                # on some occassions there are mislabels in the data, which
                # I will ignore. e.g. European 2-B Union 2-E directives 3-E
                merged_entity.update({"end": tag["end"]})
                yield merged_entity
                merged_entity = None
        elif tag["label"][-2:] == "O":
            merged_entity = tag.copy()
            yield merged_entity
            merged_entity = None


def create_train_test(
    NE_path,
    vocab_path,
    docs_path,
    train_processed_path,
    test_processed_path,
    n_sample,
    prop_train,
    rand_seed=42,
):
    """
    Function to transform the raw WiNER datasets into training
    and testing datasets suitable for training a NER model.
    This NE_path file gives the entities, but they need to be linked
    to the sentences they came from, which involves translating
    them from the vectorised form in the docs_path file
    using the vocab_path file.

    NE_path: File path to the WiNER raw data named entities
    vocab_path: File path to the WiNER raw data vocab dictionary
    docs_path: File path to the WiNER raw data documents texts
            (words coded according to the vocab)
    train_processed_path: Output file path for the training data
    test_processed_path: Output file path for the testing data
    n_sample: Number of files to get data from, each file contains about 1000 documents
    prop_train: Proportion of the file sample that will be in the training data

    Utilise the three datasources to get the entities from a sample
    of documents and save to a training and testing dataset files
    in the form:

    ID 12345

    John Smith PERSON
    rejects O
    German LOC
    call O

    Next 0
    sentence 0

    ID 444

    Next
    document
    """

    # Set the random seed for taking a sample
    random.seed(rand_seed)
    n_train = round(prop_train * n_sample)

    # Make a dictionary for the entities for every article
    # in the form
    # entities = {article_id: {sentence_id:[[0, 4, 0], [18, 25, 1]]}}
    # The sentence id is which number sentence (e.g. 1st, 2nd) the entities
    # are from in the whole document

    logger.info("Creating entities dictionary from {}".format(NE_path))
    entities = {}
    with tarfile.open(NE_path, "r:bz2") as tar:
        tar_names = tar.getnames()
        # Some nuances of the tarred folders contains
        tar_names = [
            name
            for name in tar_names
            if "CoarseNE/" in name and "CoarseNE/._" not in name
        ]
        for tar_name in tqdm(tar_names):
            member = tar.getmember(tar_name)
            f = tar.extractfile(member)
            sentence_entities = {}
            for (
                article_id,
                sentence_id,
                begin,
                end,
                ent_type,
            ) in yield_article_entities(f):
                if entities.get(article_id):
                    if entities[article_id].get(sentence_id):
                        entities[article_id][sentence_id].append([begin, end, ent_type])
                    else:
                        entities[article_id][sentence_id] = [[begin, end, ent_type]]
                else:
                    entities[article_id] = {sentence_id: [[begin, end, ent_type]]}

    # Create id2word from vocab file
    # The WiNER documentation states that the line number is the word id
    logger.info("Creating id2word dictionary from {}".format(vocab_path))
    id2word = {}
    with open(vocab_path, "r") as vocab:
        for i, line in enumerate(vocab):
            id2word[i] = line.split(" ")[0]

    # Go into a sample of the vectorised document files, see if there are any entities
    # found for it in the 'entities' dictionary, and if so translate
    # the vectorised document into words using 'id2word'
    # Saving a certain proportion to the test file and train file.
    logger.info("Creating token and tag from {}".format(docs_path))
    with tarfile.open(docs_path, "r:bz2") as tar_docs, open(
        train_processed_path, "w"
    ) as train_file, open(test_processed_path, "w") as test_file:
        tar_names = tar_docs.getnames()
        # Some nuances of the tarred folders contains
        tar_names = [
            name
            for name in tar_names
            if "Documents/Documents/" in name and "Documents/Documents/._" not in name
        ]
        # Take a sample of the relevant files
        tar_name_sample = random.sample(tar_names, n_sample)
        for i, tar_name in tqdm(enumerate(tar_name_sample)):
            member = tar_docs.getmember(tar_name)
            # Which output file to save the results in
            if i < n_train:
                output_file = train_file
            else:
                output_file = test_file
            # Extract the texts for this file if there is any given
            f = tar_docs.extractfile(member)
            if f:
                content = f.read().decode("utf-8", errors="ignore")
                # Each article in the document is separated by the article ID, e.g. 'ID 123'
                articles = content.replace("\t", " ").split("ID ")
                articles = articles[1:]
                for article in articles:
                    sentences = article.split("\n")
                    article_id = sentences[0]
                    article_entities = entities.get(article_id)
                    # Only save entity information if we found that
                    # this document indeed had entities!
                    if article_entities:
                        output_file.write("ID " + article_id)
                        output_file.write("\n\n")
                        # The first and last sentences can be blank due to splitting
                        if sentences[-1] == "":
                            sentences = sentences[1:-1]
                        else:
                            sentences = sentences[1:]
                        # In the documents each sentence is a string of
                        # numbers corresponding to words in id2word, convert
                        # this string into a list of numbers for each sentence
                        sent_wordidx = [sentence.split(" ") for sentence in sentences]
                        # Go through each of the entites for this article
                        # and translate the word ids to their words to create
                        # the tokens and each entity.
                        for sentid, sentence_entities in article_entities.items():
                            if len(sent_wordidx) > int(sentid):
                                # Translate the list of words ids for this sentence into
                                # the words using id2word
                                tokens = [
                                    id2word[int(s)] for s in sent_wordidx[int(sentid)]
                                ]
                                # Every word in the sentence is classed as an outside word
                                # until found otherwise from the entity list
                                tags = ["O"] * len(tokens)
                                for begin, end, ent_type in sentence_entities:
                                    # For entities that spread over multiple words
                                    # it should be clear which word is the start of the
                                    # entity and which is the end
                                    tags[(begin + 1):(end - 1)] = [
                                        str(ent_type) + "-I"
                                    ] * (end - 1 - begin - 1)
                                    tags[end - 1] = str(ent_type) + "-E"
                                    tags[begin] = str(ent_type) + "-B"
                                for token, tag in list(zip(tokens, tags)):
                                    output_file.write(" ".join([token, tag]))
                                    output_file.write("\n")
                            output_file.write("\n")


def _load_data_spacy(data_path, inc_outside=True, merge_entities=True):

    # Load data in Spacy format:
    # X = list of sentences (plural) / documents ['the cat ...', 'some dog...', ...]
    # Y = list of list of entity tags for each sentence
    #       [[{'start': 36, 'end': 46, 'label': 'PERSON'}, {..}, ..], ... ]
    # inc_outside = False: don't include none-entities in the output
    # merge_entities if entities span over multiple tags do you want to merge them or not
    # e.g. John 3-B Smith 3-E -> John Smith 3

    X = []
    Y = []
    sentence_text = None
    sentence_tags = None
    ent_start_index = (
        0  # A counter to populate the start and end character indexes for each entity
    )
    with open(data_path) as f:
        for line in f:
            line = line.replace("\n", "")
            if line == "" or line[0:2] == "ID":
                # You are at the start of a new sentence
                # so output previous sentence (if you had one)
                # and start variables again
                if sentence_text:
                    if merge_entities:
                        sentence_tags = [
                            merged_entity
                            for merged_entity in yield_merged_entities(
                                sentence_tags.copy()
                            )
                        ]
                    X.append(sentence_text)
                    Y.append(sentence_tags)
                sentence_text = ""
                sentence_tags = []
                ent_start_index = 0  # reset char counter for the next sentence
            else:
                token, tag = line.split(" ")
                # Add to the sentence
                sentence_text += token + " "
                if tag != "O" or inc_outside:
                    sentence_tags.append(
                        {
                            "start": ent_start_index,
                            "end": ent_start_index + len(token),
                            "label": tag,
                        }
                    )
                ent_start_index += len(token) + 1  # plus 1 for the space separating

    return X, Y


def load_winer(split="train", shuffle=True, inc_outside=True, merge_entities=True):
    path = check_cache_and_download("winer")

    if split == "train":
        train_data_path = os.path.join(path, "train.txt")
        X, Y = _load_data_spacy(
            train_data_path, inc_outside=inc_outside, merge_entities=merge_entities
        )
    elif split == "test":
        test_data_path = os.path.join(path, "test.txt")
        X, Y = _load_data_spacy(
            test_data_path, inc_outside=inc_outside, merge_entities=merge_entities
        )
    else:
        raise ValueError(
            f"Split argument {split} is not one of train, test or evaluate"
        )

    if shuffle:
        data = list(zip(X, Y))
        random.shuffle(data)
        X, Y = zip(*data)

    return X, Y


if __name__ == "__main__":

    path = check_cache_and_download("winer")

    train_processed_path = os.path.join(path, "train.txt")
    test_processed_path = os.path.join(path, "test.txt")

    if not os.path.exists(train_processed_path):
        # Since this has been done once it shouldnt need to be done again,
        # including here for completeness or in th case we want to increase
        # the sample size
        logger.info(
            "No {} training data file found, generating ...".format(
                train_processed_path
            )
        )

        NE_path = os.path.join(path, "CoarseNE.tar.bz2")
        docs_path = os.path.join(path, "Documents.tar.bz2")
        vocab_path = os.path.join(path, "document.vocab")

        # Create the train/test data
        n_sample = 10
        prop_train = 0.7

        create_train_test(
            NE_path,
            vocab_path,
            docs_path,
            train_processed_path,
            test_processed_path,
            n_sample,
            prop_train,
            rand_seed=42,
        )
    else:
        logger.info("Training and test data found")



"""Doc2Vec sklearn wrapper"""
from pathlib import Path
import multiprocessing
import statistics
import logging

from sklearn.base import BaseEstimator, TransformerMixin
from gensim.models.doc2vec import Doc2Vec, TaggedDocument
import numpy as np

logging.getLogger("gensim").setLevel(logging.WARNING)


class Doc2VecVectorizer(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        vector_size=100,
        window_size=5,
        n_jobs=1,
        min_count=2,
        negative=5,
        sample=1e-5,
        epochs=20,
        learning_rate=0.025,
        model="dm",
        pretrained=None,
    ):
        """
        Args:
            vector_size: size of vector to represent text
            window_size: words left and right of context words used to create representation
            min_count: filter words that appear less than min_count. default: 2
            negative: number of negative words to be used for training.
                      if zero hierarchical softmax is used. default: 5
            sample: threshold for downsampling high frequency words. default: 1e-5
            learning_rate: learning rate used by SGD. default: 0.025
            model: underlying model architecture, one of dm or dbow. default: dm
            epochs: number of passes over training data. default: 20
            n_jobs: number of cores to use (-1 for all). default: 1
            pretrained: path to directory containing saved pretrained doc2vec artifacts
        """
        self.vector_size = vector_size
        self.window_size = window_size
        self.epochs = epochs
        self.min_count = min_count
        self.negative = negative
        self.sample = sample
        self.n_jobs = n_jobs
        self.learning_rate = learning_rate
        self.model = model
        self.pretrained = pretrained

    def _tokenize(self, x):
        return x.lower().split()

    def _yield_tagged_documents(self, X):
        for i, x in enumerate(X):
            yield TaggedDocument(self._tokenize(x), [i])

    def fit(self, X, *_):
        """
        Args:
            X: list of texts (strings)
        """
        # If pretrained, just load, no need to fit
        if self.pretrained:
            self.load(self.pretrained)
            return self

        if self.n_jobs == -1:
            workers = multiprocessing.cpu_count()
        else:
            workers = self.n_jobs
        # TODO: Debug streaming implementation below
        #    atm it gives different result than non streaming

        # tagged_documents = self._yield_tagged_documents(X)

        # self.model = Doc2Vec(
        #    vector_size=self.vector_size, window_size=self.window_size,
        #    workers=workers, min_count=self.min_count, epochs=self.epochs
        # )
        # self.model.build_vocab(tagged_documents)
        # self.model.train(tagged_documents, total_examples=self.model.corpus_count,
        #                  epochs=self.model.epochs)
        tagged_documents = list(self._yield_tagged_documents(X))
        self.model = Doc2Vec(
            tagged_documents,
            vector_size=self.vector_size,
            window_size=self.window_size,
            workers=workers,
            min_count=self.min_count,
            epochs=self.epochs,
            negative=self.negative,
            sample=self.sample,
            alpha=self.learning_rate,
            dm=1 if self.model == "dm" else 0,
            hs=1 if self.negative == 0 else 0,
        )
        return self

    def transform(self, X):
        """
        Args:
            X: list of texts (strings)
        Returns:
            docvectors: matrix of size (nb_docs, vector_size)
        """
        return np.array([self.model.infer_vector(self._tokenize(x)) for x in X])

    def score(self, X):
        """
        Args:
            X: list of texts (strings). Needs to be the same used for fit.
        Returns:
            score: percentage of documents that are most similar with themselves
        """
        correct = []

        docvecs = self.transform(X)
        for doc_id, inferred_vector in enumerate(docvecs):
            sims = self.model.docvecs.most_similar(
                [inferred_vector], topn=len(self.model.docvecs)
            )
            rank = [docid for docid, sim in sims].index(doc_id)
            correct.append(int(rank == 0))

        return statistics.mean(correct)

    def _get_model_path(self, model_dir):
        return "{}/doc2vec".format(model_dir)

    def save(self, model_dir):
        Path(model_dir).mkdir(parents=True, exist_ok=True)
        model_path = self._get_model_path(model_dir)
        self.model.save(model_path)

    def load(self, model_dir):
        model_path = self._get_model_path(model_dir)
        self.model = Doc2Vec.load(model_path)




from nervaluate import Evaluator


def ner_classification_report(y_true, y_pred, groups, tags):
    """
    Evaluate the model's performance for each grouping of data
    for the NER labels given in 'tags'

    Input:
        y_pred: a list of predicted entities
        y_true: a list of gold entities
        groups: (str) the group each of the pred or gold entities belong to

    Output:
        report: evaluation metrics for each group
                in a nice format for printing
    """

    unique_groups = sorted(set(groups))
    outputs = []

    for group in unique_groups:
        pred_doc_entities = [y_pred[i] for i, g in enumerate(groups) if g == group]
        true_doc_entities = [y_true[i] for i, g in enumerate(groups) if g == group]

        evaluator = Evaluator(
            true_doc_entities,
            pred_doc_entities,
            tags=tags
            )
        results, _ = evaluator.evaluate()

        output_dict = {
            'precision (partial)': results['partial']['precision'],
            'recall (partial)': results['partial']['recall'],
            'f1-score': results['partial']['f1'],
            'support': len(pred_doc_entities)
            }
        output = [group]
        output.extend(list(output_dict.values()))
        outputs.append(output)

    headers = output_dict.keys()

    width = max(len(cn) for cn in unique_groups)
    head_fmt = '{:>{width}s} ' + ' {:>17}' * len(headers)
    report = head_fmt.format('', *headers, width=width)
    report += '\n\n'
    row_fmt = '{:>{width}s} ' + ' {:>17.{digits}f}' * 3 + ' {:>17}\n'

    for row in outputs:
        report += row_fmt.format(*row, width=width, digits=3)

    report += '\n'

    return report



"""
TODO: Fill this
"""
from pathlib import Path
import random

from spacy.symbols import PERSON
from spacy.tokens import Span
from spacy.util import minibatch, compounding
import spacy

from wellcomeml.ml.spacy_knowledge_base import SpacyKnowledgeBase


class SpacyEntityLinker(object):
    def __init__(self, kb_path, n_iter=50, print_output=True):
        self.kb_path = kb_path
        self.n_iter = n_iter
        self.print_output = print_output

    def _remove_examples_not_in_kb(self, kb, data):
        # Remove examples with unknown identifiers to the knowlege base
        kb_ids = kb.get_entity_strings()
        for text, annotation in data:
            for offset, kb_id_dict in annotation["links"].items():
                new_dict = {}
                for kb_id, value in kb_id_dict.items():
                    if kb_id in kb_ids:
                        new_dict[kb_id] = value
                    else:
                        print(
                            "Removed",
                            kb_id,
                            "from training because it is not in the KB.",
                        )
                annotation["links"][offset] = new_dict
        return data

    def train(self, data):
        """
        Input:
            data: list of training data in the form
                [('A sentence about Farrar',
                {'links': {(17, 22): {'Q1': 1.0, 'Q2': 0.0}}})]

        See https://spacy.io/usage/linguistic-features#entity-linking
        for where I got this code from
        """
        # TODO: Replace n_iter with self.n_iter
        n_iter = self.n_iter

        kb = SpacyKnowledgeBase()
        kb = kb.load(self.kb_path)
        print("Loaded Knowledge Base from '%s'" % self.kb_path)

        nlp = spacy.blank("en", vocab=kb.vocab)
        nlp.vocab.vectors.name = "spacy_pretrained_vectors"

        entity_linker = nlp.create_pipe("entity_linker")
        entity_linker.set_kb(kb)
        nlp.add_pipe(entity_linker, last=True)

        data = self._remove_examples_not_in_kb(kb, data)

        other_pipes = [pipe for pipe in nlp.pipe_names if pipe != "entity_linker"]
        with nlp.disable_pipes(*other_pipes):
            optimizer = nlp.begin_training()
            for itn in range(n_iter):
                random.shuffle(data)
                losses = {}
                batches = minibatch(data, size=compounding(4.0, 32.0, 1.001))
                for batch in batches:
                    texts, annotations = zip(*batch)
                    nlp.update(
                        texts, annotations, drop=0.2, losses=losses, sgd=optimizer,
                    )
                if self.print_output:
                    print(itn, "Losses", losses)

        self.nlp = nlp
        return self.nlp

    def _get_token_nums(self, doc, char_idx):
        """
        Convert a character index to a token index
        i.e. what number token is character number char_idx in ?
        """
        for i, token in enumerate(doc):
            if char_idx > token.idx:
                continue
            if char_idx == token.idx:
                return i
            if char_idx < token.idx:
                return i

    def predict(self, data):
        """
        See how well the model predicts which entity you are referring to in your data
        Input:
            data: list of test data in the form
                [('A sentence about Farrar',
                {'links': {(17, 22): {'Q1': 1.0, 'Q2': 0.0}}})]
        Output:
           pred_entities_ids: ['Q1', 'Q1', 'Q2']
    """
        # TODO: Replace nlp_el with self.nlp
        nlp_el = self.nlp

        # IMPORTANT
        #
        # This code assumes each training example has one entity
        # e.g. (start, end) = list(values['links'].keys())[0]
        #
        # If the code has more than one entities we predict the
        # first only  e.g. pred_entity_id = doc.ents[0].kb_id_

        pred_entities_ids = []
        for text, values in data:
            (start, end) = list(values["links"].keys())[0]

            doc = nlp_el.tokenizer(text)

            # Set entity span to PERSON
            doc.ents = [
                Span(
                    doc,
                    self._get_token_nums(doc, start),
                    self._get_token_nums(doc, end),
                    label=PERSON,
                )
            ]

            doc = nlp_el.get_pipe("entity_linker")(doc)

            pred_entity_id = doc.ents[0].kb_id_
            pred_entities_ids.append(pred_entity_id)

        return pred_entities_ids

    def save(self, output_dir):
        output_dir = Path(output_dir)
        if not output_dir.exists():
            output_dir.mkdir()
        self.nlp.to_disk(output_dir)
        print("Saved model to", output_dir)

    def load(self, output_dir):
        print("Loading from", output_dir)
        self.nlp = spacy.load(output_dir)
        return self.nlp



#!/usr/bin/env python3
# coding: utf-8

"""
A generic "frequency" vectorizer that wraps all usual transformations.
"""
import re

import numpy as np
import spacy
from sklearn.feature_extraction.text import TfidfVectorizer

from wellcomeml.logger import logger


class WellcomeTfidf(TfidfVectorizer):
    """
    Class to wrap some basic transformation and text
    vectorisation/embedding
    """

    def __init__(self, **kwargs):
        """

        Args:
            Any sklearn "tfidfvectorizer" arguments (min_df, etc.)

        """
        self.embedding = "tf-idf"

        logger.info("Initialising frequency vectorizer.")

        kwargs["stop_words"] = kwargs.get("stop_words", "english")

        super().__init__(**kwargs)

    @classmethod
    def save_transformed(cls, path, X_transformed):
        """Saves transformed embedded vectors"""
        # Numpy will save the transformed sparse array
        # into a 0-dimensional array, preserving sparsity
        np.save(path, X_transformed)

    @classmethod
    def load_transformed(cls, path):
        """Loads transformed embedded vectors"""

        # Loads the 0-dimensional array and returns its content
        X_transformed = np.load(path, allow_pickle=True)

        return X_transformed[()]

    def regex_transform(self, X, remove_numbers="years", *_):
        """
        Extra regular expression transformations to clean text
        Args:
            X: A list of texts (strings)
            *_:
            remove_numbers: Whether to remove years or all digits. Caveat:
            This does not only remove years, but **any number** between
            1000 and 2999.

        Returns:
            A list of texts with the applied regex transformation

        """
        if remove_numbers == "years":
            return [re.sub(r"[1-2]\d{3}", "", text) for text in X]
        elif remove_numbers == "digits":
            return [re.sub(r"\d", "", text) for text in X]
        else:
            return X

    def spacy_lemmatizer(self, X, remove_stopwords_and_punct=True):
        """
        Uses spacy pre-trained lemmatisation model to
        Args:
            X: A list of texts (strings)
            remove_stopwords_and_punct: Whether to remove stopwords,
            punctuation, pronouns

        Returns:

        """

        try:
            nlp = spacy.load(
                "en_core_web_sm", disable=["ner", "tagger", "parser", "textcat"]
            )
        except IOError:
            from wellcomeml.__main__ import download

            download("models")
            # pkg_resources need to be reloaded to pick up the newly installed models
            import pkg_resources
            import imp

            imp.reload(pkg_resources)
            nlp = spacy.load(
                "en_core_web_sm", disable=["ner", "tagger", "parser", "textcat"]
            )

        logger.info("Using spacy pre-trained lemmatiser.")
        if remove_stopwords_and_punct:
            return [
                [
                    token.lemma_.lower()
                    for token in doc
                    if not token.is_stop
                    and not token.is_punct
                    and token.lemma_ != "-PRON-"
                ]
                for doc in nlp.tokenizer.pipe(X)
            ]
        else:
            return [
                [token.lemma_.lower() for token in doc] for doc in nlp.tokenizer.pipe(X)
            ]

    def transform(self, X, regex=True, spacy_lemmatizer=True, *_):
        if regex:
            X = self.regex_transform(X)
        if spacy_lemmatizer:
            X = self.spacy_lemmatizer(X)

        X = [" ".join(text) for text in X]

        return super().transform(X)

    def fit(self, X, regex=True, spacy_lemmatizer=True, *_):
        if regex:
            X = self.regex_transform(X)
        if spacy_lemmatizer:
            X = self.spacy_lemmatizer(X)

        logger.info("Fitting vectorizer.")

        X = [" ".join(text) for text in X]
        super().fit(X)

        return self



import os
from wellcomeml.logger import logger

# Introduced a development_transformers env variable, that allows to
# disable functions that use spacy.

development_transformers_mode = (
    os.environ.get("WELLCOMEML_ENV", "") == "development_transformers"
)

if development_transformers_mode:
    logger.warning(
        "Running in development mode. Only loading modules that"
        " use new version of transformers."
    )

    from .bert_semantic_equivalence import SemanticEquivalenceClassifier
    __all__ = [SemanticEquivalenceClassifier]
else:
    from .frequency_vectorizer import WellcomeTfidf
    from .doc2vec_vectorizer import Doc2VecVectorizer
    from .sent2vec_vectorizer import Sent2VecVectorizer
    from .voting_classifier import WellcomeVotingClassifier
    __all__ = [WellcomeTfidf, Doc2VecVectorizer,
               Sent2VecVectorizer, WellcomeVotingClassifier]

    try:
        from .vectorizer import Vectorizer
        from .clustering import TextClustering
        from .spacy_ner import SpacyNER
        from .spacy_classifier import SpacyClassifier
        from .bert_classifier import BertClassifier
        from .bert_vectorizer import BertVectorizer
        from .spacy_knowledge_base import SpacyKnowledgeBase
        from .spacy_entity_linking import SpacyEntityLinker
        from .similarity_entity_linking import SimilarityEntityLinker
        from .bert_semantic_equivalence import SemanticEquivalenceClassifier
        from .cnn import CNNClassifier
        from .bilstm import BiLSTMClassifier
        from .keras_vectorizer import KerasVectorizer
        __all__ += [
            Vectorizer, TextClustering, SpacyNER, SpacyClassifier,
            BertClassifier, BertVectorizer, SpacyKnowledgeBase,
            SpacyEntityLinker, SemanticEquivalenceClassifier, CNNClassifier,
            BiLSTMClassifier, KerasVectorizer, SimilarityEntityLinker
        ]
    except ImportError:
        logger.warning("Using WellcomeML without extras (transformers & torch).")



from typing import Tuple
from collections import OrderedDict


class WellcomeColor(object):
    """
    Represents a Wellcome named color
    """

    def __init__(self,
                 name: str,
                 hex: str,
                 rgb: Tuple[int, int, int]):
        """
        Specifies a named color

        Args:
            hex: Hex representation
            name: Name of the color
            rgb: Tuple of 3 ints specifying RGB representation
        """
        self.hex = hex
        self.name = name
        self.rgb = rgb

    def __repr__(self) -> str:
        """The representation of a color is its hex"""
        return self.hex

# Name colors as in https://company-57536.frontify.com/d/
# gFEfjydViLRJ/wellcome-brand-book#/visuals/dataviz-elements-and-rationale


NAMED_COLORS_DICT = OrderedDict((
    ("Blue Lagoon",
     WellcomeColor("Blue Lagoon", "#008094", (0, 128, 148))),
    ("Amber",
     WellcomeColor("Amber", "#FEC200", (254, 194, 0))),
    ("Tahiti Gold",
     WellcomeColor("Tahiti Gold", "#F07F09", (240, 127, 9))),
    ("Brick Red",
     WellcomeColor("Brick Red", "#CF334D", (207, 51, 77))),
    ("Midnight Blue",
     WellcomeColor("Midnight Blue", "#003170", (0, 49, 112))),
    ("Rope",
     WellcomeColor("Rope", "#8A471E", (138, 71, 30))),
    ("Eunry",
     WellcomeColor("Eunry", "#CD9789", (205, 151, 137))),
    ("Fountain Blue",
     WellcomeColor("Fountain Blue", "#62C0CE", (98, 192, 206))),
    ("Sea Pink",
     WellcomeColor("Sea Pink", "#ED858E", (237, 133, 142))),
    ("Mantis",
     WellcomeColor("Mantis", "#90C879", (144, 200, 121))),
    ("Olive Drab",
     WellcomeColor("Olive Drab", "#4C8026", (76, 128, 38)))
))

# Named colors, interpolated with 30% decrements to generate a bigger palette
NAMED_COLORS_LARGE_DICT = OrderedDict((
    ("Blue Lagoon",
     WellcomeColor("Blue Lagoon", "#008094", (0, 128, 148))),
    ("Blue Lagoon 2",
     WellcomeColor("Blue Lagoon 2", "#005967", (0, 89, 103))),
    ("Blue Lagoon 3",
     WellcomeColor("Blue Lagoon 3", "#003e48", (0, 62, 72))),
    ("Amber",
     WellcomeColor("Amber", "#FEC200", (254, 194, 0))),
    ("Amber 2",
     WellcomeColor("Amber 2", "#b18700", (177, 135, 0))),
    ("Amber 3",
     WellcomeColor("Amber 3", "#7b5e00", (123, 94, 0))),
    ("Tahiti Gold",
     WellcomeColor("Tahiti Gold", "#F07F09", (240, 127, 9))),
    ("Tahiti Gold 2",
     WellcomeColor("Tahiti Gold 2", "#a85806", (168, 88, 6))),
    ("Tahiti Gold 3",
     WellcomeColor("Tahiti Gold 3", "#753d04", (117, 61, 4))),
    ("Brick Red",
     WellcomeColor("Brick Red", "#CF334D", (207, 51, 77))),
    ("Brick Red 2",
     WellcomeColor("Brick Red 2", "#902335", (144, 35, 53))),
    ("Brick Red 3",
     WellcomeColor("Brick Red 3", "#641825", (100, 24, 37))),
    ("Midnight Blue",
     WellcomeColor("Midnight Blue", "#003170", (0, 49, 112))),
    ("Midnight Blue 2",
     WellcomeColor("Midnight Blue 2", "#00224e", (0, 34, 78))),
    ("Midnight Blue 3",
     WellcomeColor("Midnight Blue 3", "#001736", (0, 23, 54))),
    ("Rope",
     WellcomeColor("Rope", "#8A471E", (138, 71, 30))),
    ("Rope 2",
     WellcomeColor("Rope 2", "#603115", (96, 49, 21))),
    ("Rope 3",
     WellcomeColor("Rope 3", "#43220e", (67, 34, 14))),
    ("Eunry",
     WellcomeColor("Eunry", "#CD9789", (205, 151, 137))),
    ("Eunry 2",
     WellcomeColor("Eunry 2", "#8f695f", (143, 105, 95))),
    ("Eunry 3",
     WellcomeColor("Eunry 3", "#644942", (100, 73, 66))),
    ("Fountain Blue",
     WellcomeColor("Fountain Blue", "#62C0CE", (98, 192, 206))),
    ("Fountain Blue 2",
     WellcomeColor("Fountain Blue 2", "#448690", (68, 134, 144))),
    ("Fountain Blue 3",
     WellcomeColor("Fountain Blue 3", "#2f5d64", (47, 93, 100))),
    ("Sea Pink",
     WellcomeColor("Sea Pink", "#ED858E", (237, 133, 142))),
    ("Sea Pink 2",
     WellcomeColor("Sea Pink 2", "#a55d63", (165, 93, 99))),
    ("Sea Pink 3",
     WellcomeColor("Sea Pink 3", "#734145", (115, 65, 69))),
    ("Mantis",
     WellcomeColor("Mantis", "#90C879", (144, 200, 121))),
    ("Mantis 2",
     WellcomeColor("Mantis 2", "#648c54", (100, 140, 84))),
    ("Mantis 3",
     WellcomeColor("Mantis 3", "#46623a", (70, 98, 58))),
    ("Olive Drab",
     WellcomeColor("Olive Drab", "#4C8026", (76, 128, 38))),
    ("Olive Drab 2",
     WellcomeColor("Olive Drab 2", "#35591a", (53, 89, 26))),
    ("Olive Drab 3",
     WellcomeColor("Olive Drab 3", "#253e12", (37, 62, 18)))
))



#!/usr/bin/env python3
# coding: utf-8

"""
Set up shared logger
"""

import logging
import os

LOGGING_LEVEL = os.getenv("LOGGING_LEVEL")

if isinstance(LOGGING_LEVEL, str):
    numeric_level = getattr(logging, LOGGING_LEVEL.upper(), 10)
else:
    numeric_level = 20

logger = logging.getLogger(__name__)

logging.basicConfig(
    format="%(asctime)s %(name)s %(levelname)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=numeric_level,
)

external_logging_level = {'transformers': logging.WARNING}

for package, level in external_logging_level.items():
    logging.getLogger(package).setLevel(level)



# encoding: utf-8
"""
BERT Vectorizer that embeds text using a prertained BERT model
"""

from transformers import BertModel, BertTokenizer
from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
import torch

from wellcomeml.utils import check_cache_and_download
from wellcomeml.logger import logger


class BertVectorizer(BaseEstimator, TransformerMixin):
    def __init__(self, pretrained="bert", sentence_embedding="mean_second_to_last"):
        """
        Bert vectorizer parameters

        Args:
            pretrained: A pre-trained model name. Currently 'bert' or 'scibert'
            sentence_embedding: How to embedd a sentence using bert's layers.
            Current options:
            'mean_second_to_last', 'mean_last', 'sum_last' or 'mean_last_four'
            Default: `mean_second_to_last`. If a valid option is not set,
            returns the pooler layer (embedding for the token [CLS])
        """
        self.pretrained = pretrained
        self.sentence_embedding = sentence_embedding

    @classmethod
    def save_transformed(cls, path, X_transformed):
        """Saves transformed embedded vectors"""
        np.save(path, X_transformed)

    @classmethod
    def load_transformed(cls, path):
        """Loads transformed embedded vectors"""
        return np.load(path)

    def bert_embedding(self, x):
        # Max sequence length is 512 for BERT
        if len(x) > 512:
            embedded_a = self.bert_embedding(x[:512])
            embedded_b = self.bert_embedding(x[512:])
            return embedded_a + embedded_b

        tokenized_x = self.tokenizer.tokenize("[CLS] " + x + " [SEP]")
        indexed_tokens = self.tokenizer.convert_tokens_to_ids(tokenized_x)

        tokens_tensor = torch.tensor([indexed_tokens])
        segments_tensor = torch.zeros(tokens_tensor.shape, dtype=torch.long)
        with torch.no_grad():
            output = self.model(tokens_tensor, token_type_ids=segments_tensor)
        last_layer = output[2][-1]
        second_to_last_layer = output[2][-2]

        if self.sentence_embedding == "mean_second_to_last":
            embedded_x = second_to_last_layer.mean(dim=1)
        elif self.sentence_embedding == "mean_last":
            embedded_x = last_layer.mean(dim=1)
        elif self.sentence_embedding == "sum_last":
            embedded_x = last_layer.sum(dim=1)
        elif self.sentence_embedding == "mean_last_four":
            embedded_x = torch.stack(output[2][-4:]).mean(dim=0).mean(dim=1)
        else:
            # Else gives the embbedding for the pooler layer. This can be, for
            # example, fed into a softmax classifier
            embedded_x = output[1]

        return embedded_x.cpu().numpy().flatten()

    def transform(self, X, *_):
        return np.array([self.bert_embedding(x) for x in X])

    def fit(self, *_):
        model_name = (
            "bert-base-uncased"
            if self.pretrained == "bert"
            else "scibert_scivocab_uncased"
        )

        # If model_name doesn't exist checks cache and change name to
        # full path
        if model_name == "scibert_scivocab_uncased":
            model_name = check_cache_and_download(model_name)

        logger.info("Using {} embedding".format(model_name))
        self.model = BertModel.from_pretrained(model_name, output_hidden_states=True)
        self.tokenizer = BertTokenizer.from_pretrained(model_name)
        self.model.eval()
        return self



import tarfile
import os

import boto3
from botocore import UNSIGNED
from botocore.client import Config

from wellcomeml.logger import logger

MODELS_DIR = os.path.expanduser("~/.cache/wellcomeml/models")

MODEL_DISPATCH = {
    'scibert_scivocab_uncased': {
        "bucket": "ai2-s2-research",
        "path": "scibert/huggingface_pytorch/scibert_scivocab_uncased.tar",
        "file_name": "scibert_scivocab_uncased.tar"
    },
    'scibert_scivocab_cased': {
        "bucket": "ai2-s2-research",
        "path": "scibert/huggingface_pytorch/scibert_scivocab_cased.tar",
        "file_name": "scibert_scivocab_cased.tar"
    },
    'biosent2vec': {
        "bucket": "datalabs-public",
        "path": "models/ncbi-nlp/biosent2vec.bin",
        "file_name": "biosent2vec.bin",
    },
    'sent2vec_wiki_unigrams': {
        "bucket": "datalabs-public",
        "path": "models/epfml/wiki_unigrams.bin",
        "file_name": "wiki_unigrams.bin",
    }
}


def check_cache_and_download(model_name):
    """ Checks if model_name is cached and return complete path"""
    os.makedirs(MODELS_DIR, exist_ok=True)

    FILE_NAME = MODEL_DISPATCH[model_name]['file_name']
    _, FILE_EXT = FILE_NAME.split(".")

    model_path = os.path.join(MODELS_DIR, model_name if FILE_EXT == "tar" else FILE_NAME)
    if not os.path.exists(model_path):
        logger.info(f"Could not find model {model_name}. Downloading from S3")

        # The following allows to download from S3 without AWS credentials
        s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
        tmp_file = os.path.join(MODELS_DIR, FILE_NAME)

        s3.download_file(MODEL_DISPATCH[model_name]['bucket'],
                         MODEL_DISPATCH[model_name]['path'], tmp_file)

        if FILE_EXT == 'tar':
            tar = tarfile.open(tmp_file)
            tar.extractall(path=MODELS_DIR)
            tar.close()

            os.remove(tmp_file)

    return model_path



from collections import defaultdict

from sklearn.metrics import f1_score, precision_score, recall_score
import tensorflow as tf


class Metrics(tf.keras.callbacks.Callback):
    def __init__(self, validation_data):
        self.validation_data = validation_data

    def on_epoch_end(self, epoch, logs):
        X_val = self.validation_data[0]
        Y_val = self.validation_data[1]

        Y_pred = self.model.predict(X_val) > 0.5
        f1 = round(f1_score(Y_val, Y_pred, average="micro"), 4)
        p = round(precision_score(Y_val, Y_pred, average="micro"), 4)
        r = round(recall_score(Y_val, Y_pred, average="micro"), 4)
        print(f" - val metrics: P {p:.4f} R {r:.4f} F {f1:.4f}")
        return


class CategoricalMetrics(tf.keras.metrics.Metric):
    def __init__(
        self,
        metric="precision",
        from_logits=True,
        threshold=0.5,
        binary="true",
        pos=1,
        **kwargs,
    ):
        """
        Categorical metrics

        Args:
            metric: 'precision', 'recall' or 'f1'
            from_logits: Similar to keras 'from_logits'. If True, calculates
                        a softmax activation on input
            threshold: a threshold to calculate predictions
            binary: whether input is binary or not (in which case does average)
            pos: the positive label (between [0, n_classes])
            **kwargs:
        """
        self.metric = metric
        name = f"categorical_{metric}"
        super().__init__(name=name, **kwargs)
        self.threshold = threshold
        self.from_logits = from_logits
        self.binary = binary
        self.pos = pos
        self.value = self.add_weight(name="tp", initializer="zeros")

    def update_state(self, y_true, y_pred, sample_weight=None):
        if self.from_logits:
            y_pred = tf.keras.activations.softmax(y_pred)

        greater_than_threshold = tf.cast(y_pred[:, self.pos:] > self.threshold, "bool")
        positive = tf.cast(y_true, "int32") == self.pos

        # Epsilon added to denominators to avoid division by zero
        eps = tf.keras.backend.epsilon()

        tp = tf.reduce_sum(
            tf.cast(tf.logical_and(greater_than_threshold, positive), "float32")
        )

        if self.metric in ["precision", "f1_score"]:
            tp_plus_fp = tf.reduce_sum(tf.cast(greater_than_threshold, "float32"))
            precision = tp / (tp_plus_fp + eps)

        if self.metric in ["recall", "f1_score"]:
            tp_plus_fn = tf.reduce_sum(tf.cast(positive, "float32"))
            recall = tp / (tp_plus_fn + eps)

        if self.metric == "precision":
            self.value.assign(precision)
        elif self.metric == "recall":
            self.value.assign(recall)
        elif self.metric == "f1_score":
            self.value.assign(2 * precision * recall / (precision + recall + eps))

    def result(self):
        return self.value

    def reset_states(self):
        # The state of the metric will be reset at the start of each epoch.
        self.value.assign(0.0)


class MetricMiniBatchHistory(tf.keras.callbacks.Callback):
    def on_train_begin(self, logs={}):
        self.metric_history_mini_batch = defaultdict(list)

    def on_batch_end(self, batch, logs={}):
        for metric in self.params["metrics"]:
            if logs.get(metric):
                self.metric_history_mini_batch[metric].append(logs.get(metric))
                self.model.history.history[
                    f"mini_batch_{metric}"
                ] = self.metric_history_mini_batch[metric]
